<?php
// $Id: install.oracle.inc,v 1.1.2.7 2009/12/03 00:02:45 andreagariboldi Exp $

define('DRUPAL_ROOT', getcwd());
include_once DRUPAL_ROOT.'/includes/database.oracle.inc';

// Oracle specific install functions
openlog("database.oracle.log", LOG_PID | LOG_PERROR, LOG_USER);
/**
 * Check if Oracle is available.
 *
 * @return
 *  TRUE/FALSE
 */
function oracle_is_available() {
  foreach (PDO::getAvailableDrivers() as $driver)
     if ($driver=='oci')
       return true; 
	
  return false;
}

/**
 * Check if we can connect to Oracle.
 *
 * @return
 *  TRUE/FALSE
 */
function drupal_test_oracle($url, &$success)
{
	  global $active_db;
	  
	  if (!oracle_is_available())
	  {
	    drupal_set_message(st('PHP Oracle support not enabled.'), 'error');
	    return false;
	  }
	
	  $url = parse_url($url);
	  $conn_string = '';
	
	  $url['user'] = urldecode($url['user']);
	  $url['pass'] = isset($url['pass']) ? urldecode($url['pass']) : '';
	  $url['host'] = urldecode($url['host']);
	  $url['path'] = urldecode($url['path']);
	
	  if (isset($url['port'])) 
	    $url['host'] = $url['host'] .':'. $url['port'];
	  
	
	  $url['host'] = "//" . $url['host'] . $url['path']; 
	
	  $dsn = 'oci:dbname=' . $url['host'] . ';charset=AL32UTF8';
	  syslog(LOG_ERR,"dsn: ".$dsn);
	  $conn= false;
	  $success = array();
	  
	  try
	  {
		  $conn= new PDO($dsn, $url['user'], $url['pass']);
	
		  $conn->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);
		  $conn->setAttribute(PDO::ATTR_CASE, PDO::CASE_LOWER);
		  $conn->setAttribute(PDO::ATTR_STRINGIFY_FETCHES, TRUE);
		  
		  $active_db= $conn;
          $success[] = 'CONNECT';
	  }
	  catch (Exception $ex)
	  {
          drupal_set_message(st('Failed to connect to your Oracle database server. Oracle reports the following message: %error.<ul><li>Are you sure you have the correct username and password?</li><li>Are you sure that you have typed the correct database hostname?</li><li>Are you sure that the database server is running?</li><li>Are you sure you typed the correct database name?</li></ul>For more help, see the <a href="http://drupal.org/node/258">Installation and upgrading handbook</a>. If you are unsure what these terms mean you should probably contact your hosting provider.', array('%error' => $ex->getMessage())), 'error');
          return false;
	  }
	  
	  try
	  {
		  // Test CREATE.
		  
	  	  $query= 'CREATE TABLE drupal_install_test (id number NOT NULL)';
	  	  
		  oracle_query($query);
		  
		  $success[] = 'CREATE';
	  }
	  catch (Exception $ex)
	  {
          drupal_set_message(st('Failed to create a test table on your Oracle database server with the command %query. Oracle reports the following message: %error.<ul><li>Are you sure the configured username has the necessary Oracle permissions to create tables in the database?</li></ul>For more help, see the <a href="http://drupal.org/node/258">Installation and upgrading handbook</a>. If you are unsure what these terms mean you should probably contact your hosting provider.', array('%query' => $query, '%error' => $ex->getMessage())), 'error');
          return false;
	  }
	  
	  $active_db->beginTransaction();
	  
	  try
	  {
		  // Test INSERT.
		  
	  	  $query= 'INSERT INTO drupal_install_test (id) VALUES (1)';
	  	  
		  oracle_query($query);
		  
		  $success[] = 'INSERT';
	  }
	  catch (Exception $ex)
	  {
          drupal_set_message(st('Failed to insert a value into a test table on your Oracle database server. We tried inserting a value with the command %query and Oracle reported the following error: %error.', array('%query' => $query, '%error' => $ex->getMessage())), 'error');
          return false;
	  }
	  
	  try
	  {
		  // Test SELECT.
		  
	  	  $query= 'select * from drupal_install_test';
	  	  
		  $obj= oracle_query($query)->fetchObject();
		  
		  $success[] = 'SELECT';
	  }
	  catch (Exception $ex)
	  {
          drupal_set_message(st('Failed to select a test table on your Oracle database server with the command %query. Oracle reports the following message: %error.<ul><li>Are you sure the configured username has the necessary Oracle permissions to create tables in the database?</li></ul>For more help, see the <a href="http://drupal.org/node/258">Installation and upgrading handbook</a>. If you are unsure what these terms mean you should probably contact your hosting provider.', array('%query' => $query, '%error' => $ex->getMessage())), 'error');
          return false;
	  }
	  

	  try
	  {
		  // Test UPDATE.
		  
	  	  $query= 'UPDATE drupal_install_test SET id = 2';
	  	  
		  $obj= oracle_query($query);
		  
		  $success[] = 'UPDATE';
	  }
	  catch (Exception $ex)
	  {
          drupal_set_message(st('Failed to update a test table on your Oracle database server with the command %query. Oracle reports the following message: %error.<ul><li>Are you sure the configured username has the necessary Oracle permissions to create tables in the database?</li></ul>For more help, see the <a href="http://drupal.org/node/258">Installation and upgrading handbook</a>. If you are unsure what these terms mean you should probably contact your hosting provider.', array('%query' => $query, '%error' => $ex->getMessage())), 'error');
          return false;
	  }
	  
	  
	  try
	  {
		  // Test LOCK.
		  
	  	  $query= 'LOCK TABLE drupal_install_test IN SHARE ROW EXCLUSIVE MODE';
	  	  
		  $obj= oracle_query($query);
		  
		  $success[] = 'UPDATE';
	  }
	  catch (Exception $ex)
	  {
          drupal_set_message(st('Failed to lock a test table on your Oracle database server with the command %query. Oracle reports the following message: %error.<ul><li>Are you sure the configured username has the necessary Oracle permissions to create tables in the database?</li></ul>For more help, see the <a href="http://drupal.org/node/258">Installation and upgrading handbook</a>. If you are unsure what these terms mean you should probably contact your hosting provider.', array('%query' => $query, '%error' => $ex->getMessage())), 'error');
          return false;
	  }
	  
	  try
	  {
		  // Test UNLOCK.
		  
	  	  $query= 'COMMIT';
	  	  
    	  $conn->commit(); 
          $success[] = 'UNLOCK';
          
	  }
	  catch (Exception $ex)
	  {
          drupal_set_message(st('Failed to lock a test table on your Oracle database server with the command %query. Oracle reports the following message: %error.<ul><li>Are you sure the configured username has the necessary Oracle permissions to create tables in the database?</li></ul>For more help, see the <a href="http://drupal.org/node/258">Installation and upgrading handbook</a>. If you are unsure what these terms mean you should probably contact your hosting provider.', array('%query' => $query, '%error' => $ex->getMessage())), 'error');
          return false;
	  }
  

	  try
	  {
		  // Test UPDATE.
		  
	  	  $query= 'DELETE FROM drupal_install_test';
	  	  
		  $obj= oracle_query($query);
		  
		  $success[] = 'DELETE';
	  }
	  catch (Exception $ex)
	  {
          drupal_set_message(st('Failed to delete a test table on your Oracle database server with the command %query. Oracle reports the following message: %error.<ul><li>Are you sure the configured username has the necessary Oracle permissions to create tables in the database?</li></ul>For more help, see the <a href="http://drupal.org/node/258">Installation and upgrading handbook</a>. If you are unsure what these terms mean you should probably contact your hosting provider.', array('%query' => $query, '%error' => $ex->getMessage())), 'error');
          return false;
	  }
	  

	  try
	  {
		  // Test DROP.
		  
	  	  $query= 'DROP TABLE drupal_install_test';
	  	  
		  $obj= oracle_query($query);
		  
		  $success[] = 'DROP';
	  }
	  catch (Exception $ex)
	  {
          drupal_set_message(st('Failed to drop a test table on your Oracle database server with the command %query. Oracle reports the following message: %error.<ul><li>Are you sure the configured username has the necessary Oracle permissions to create tables in the database?</li></ul>For more help, see the <a href="http://drupal.org/node/258">Installation and upgrading handbook</a>. If you are unsure what these terms mean you should probably contact your hosting provider.', array('%query' => $query, '%error' => $ex->getMessage())), 'error');
          return false;
	  }
	  
	  // create driver needed objects
      oracle_failsafe_ddl("create table blobs (blobid number primary key, content blob, hash varchar2(32))"); // used to store blob values
      oracle_failsafe_ddl("create unique index idx_blobs_hash on blobs (hash)"); // used to index blob hash
      oracle_failsafe_ddl("create sequence seq_blobs"); // blobid primary key sequence
      oracle_failsafe_ddl("create table blob_column (table_name varchar2(255), column_name varchar2(255))"); // blob column catalog
      oracle_failsafe_ddl("create table long_identifiers (id number primary key, identifier varchar2(4000))"); // used to store long identifier catalog
      oracle_failsafe_ddl("create sequence seq_long_identifiers"); // long_identifiers primary key sequence
      oracle_failsafe_ddl("create table used_blobs ( blobid number primary key ) organization index");

      oracle_query("create or replace package drupal_timestamp
as

   /* this package can be used to convert drupal timestamp numbers to oracle dates and vice versa
    *
    * e.g.: select to_char(drupal_timestamp.todate(created),'DD/MM/YYYY HH24:MI:SS') created from node
    */
   
   function todate(p_timestamp number)
   return date
   deterministic;
   
   
   function totimestamp(p_date date)
   return number
   deterministic;   
      
end;");
      
      oracle_query("create or replace package body drupal_timestamp
as
   
  PHP_BASE_TIME   timestamp:= (from_tz(to_timestamp('01011970000000','DDMMYYYYHH24MISS'), '+00:00') at time zone '+00:00');
  PHP_BASE_DATE   date:= to_date(to_char(PHP_BASE_TIME at time zone '00:00','DDMMYYYYHH24MISS'),'DDMMYYYYHH24MISS') at time zone '00:00';


   function todate(p_timestamp number)
   return date
   deterministic
   as
   begin
     return ((PHP_BASE_TIME at time zone '+00:00')+p_timestamp/86400) at time zone '+00:00';
   end;
   
   
   function totimestamp(p_date date)
   return number
   deterministic
   as
   begin
     return round((p_date-PHP_BASE_DATE)*86400);
   end;
      
end;");
      
      oracle_query("create or replace package blobs_cleaner
as

  type num_table is table of number;
  
  function used_blobs
  return num_table
  pipelined;

  procedure cleanup;
  
end blobs_cleaner;");
      
      oracle_query("create or replace package body blobs_cleaner as

  type refcur is ref cursor;
  
  function used_blobs
  return num_table
  pipelined
  as
    c_ids refcur;
    v_num number;
  begin
  
    for c_cur in (select distinct table_name, column_name from blob_column) loop
       
       open c_ids for 'select distinct '||c_cur.column_name||' from '||c_cur.table_name;
       <<tab>>
       loop
          fetch c_ids into v_num;
          exit tab when c_ids%notfound;
          pipe row(v_num);
       end loop;
       close c_ids;
       
    end loop;
  
    return;
  end used_blobs;
  
  procedure cleanup
  as
  begin
  
    lock table blobs in exclusive mode;
    
    begin
     execute immediate 'truncate table used_blobs';
    exception
     when others then
       null;
    end;
    
    insert into used_blobs select distinct column_value blobid from table(blobs_cleaner.used_blobs);
    
    delete blobs a
     where not exists (select 1 from used_blobs where blobid= a.blobid);
     
    commit;
  
  end;

end blobs_cleaner;");
      
      oracle_query("create or replace package identifier
as

  long_identifier_prefix  constant varchar2(10):= '".ORACLE_LONG_IDENTIFIER_PREFIX."';
  identifier_max_length   constant number:= ".ORACLE_IDENTIFIER_MAX_LENGTH.";
  empty_replacer_char      constant varchar2(10):= '".ORACLE_EMPTY_STRING_REPLACER."';
  
  type serial_info is record
  (
      sequence_name    varchar2(30),
      field_name       varchar2(30),
      trigger_name     varchar2(30),
      sequence_restart number
  );
  
  type serial_info_tab is table of serial_info;

  function get_for(p_long_identifier varchar2)
  return varchar2;
  
  function sequence_for_table(p_table varchar2, p_schema varchar2)
  return varchar2;

  function get_serial(p_table varchar2, p_schema varchar2)
  return serial_info_tab
  pipelined;
  
  procedure write_blob(p_hash varchar2, p_blobid out number, p_blob out blob);
  
  procedure check_db_prefix(p_db_prefix varchar2);

end identifier;");      
      
      oracle_query("create or replace package body identifier
as

  function get_for(p_long_identifier varchar2)
  return varchar2
  as pragma autonomous_transaction;
   v_id number;
  begin
  
     if length(p_long_identifier) < identifier_max_length+1 then
       return p_long_identifier;
     end if;
     
     select id
       into v_id
       from long_identifiers
      where identifier= upper(p_long_identifier);
     
     return long_identifier_prefix||v_id;
  
  exception
   when no_data_found then

     insert into long_identifiers (id,identifier)
          values (seq_long_identifiers.nextval,upper(p_long_identifier))
       returning id into v_id;
     commit;

     return long_identifier_prefix||v_id;
     
  end;
  
  function sequence_for_table(p_table varchar2, p_schema varchar2)
  return varchar2
  as
    v_contraint_oname varchar2(31);
    v_table_oname     varchar2(31):= upper(get_for(p_table));
    v_col_name        varchar2(4000);
    v_sequence_oname  varchar2(31);
    v_schema          varchar2(31):= nvl(upper(p_schema),user);
  begin
     
     select constraint_name
       into v_contraint_oname
       from all_constraints
      where constraint_type= 'P'
        and table_name= v_table_oname
        and owner= v_schema;
        
     select column_name
       into v_col_name
       from all_cons_columns
      where constraint_name= v_contraint_oname
        and table_name= v_table_oname
        and owner= v_schema;
        
     if instr(v_col_name,long_identifier_prefix) > 0 then
      
       declare
         v_id number;
       begin
         v_id:= to_number(substr(v_col_name,4));
         select identifier
           into v_col_name
           from long_identifiers
          where id= v_id;
       end;
       
     end if;
        
     v_sequence_oname:= get_for('SEQ_'||upper(p_table)||'_'||v_col_name);
     
     select sequence_name
       into v_sequence_oname
       from all_sequences 
      where sequence_name= v_sequence_oname
        and sequence_owner= v_schema;
      
     return '\"'||v_schema||'\".\"'||v_sequence_oname||'\"';
        
  exception
   when others then
     return null;
  end;

  function get_serial(p_table varchar2, p_schema varchar2)
  return serial_info_tab
  pipelined
  as
    v_serial          serial_info;
    v_table_oname     varchar2(30):= upper(get_for(p_table));
    v_schema          varchar2(31):= nvl(upper(p_schema),user);
    v_serial_re       varchar2(255):= '\/\* serial\((\w+),(\w+),(\w+),(\w+)\) \*\/';
  begin
  
     select regexp_replace(s.text,v_serial_re,'\\2'),
            regexp_replace(s.text,v_serial_re,'\\3'),
            regexp_replace(s.text,v_serial_re,'\\4')
       into v_serial.trigger_name,
            v_serial.sequence_name,
            v_serial.field_name
       from (select regexp_substr(text,v_serial_re) text
               from all_source
              where text like '%/* serial('||v_table_oname||',%'
                and type= 'TRIGGER' 
                and line= 1 
                and owner= v_schema) s;

     select last_number+1
       into v_serial.sequence_restart
       from all_sequences
      where sequence_name= v_serial.sequence_name
        and sequence_owner= v_schema;       
  
     pipe row(v_serial);  -- WARN: assume 1 serial for table for now seems right
     
     return;
     
  exception
   when others then
      return;
  end;

  procedure write_blob(p_hash varchar2, p_blobid out number, p_blob out blob)
  as
  begin
    select blobid,
           content
      into p_blobid,
           p_blob
      from blobs
     where hash= p_hash;

  exception
   when no_data_found then
    insert into blobs (blobid,content,hash)
         values (seq_blobs.nextval,empty_blob,p_hash)
      returning blobid, content into p_blobid,
                                     p_blob;
  end;
  
  procedure check_db_prefix(p_db_prefix varchar2)
  as pragma autonomous_transaction;
    v_db_prefix   varchar2(30):= upper(p_db_prefix);
  begin
  
     select username
       into v_db_prefix
       from all_users
      where username= v_db_prefix;
  
  exception
    when no_data_found then
      execute immediate 'grant connect, resource to \"'||v_db_prefix||'\" identified by \"'||v_db_prefix||'\"';
  end;
    
end identifier;");     

      // compat functions

      oracle_failsafe_ddl("create or replace function greatest(p1 number, p2 number, p3 number default null)
return number 
as
begin
  if p3 is null then
    if p1 > p2 or p2 is null then
     return p1;
    else
     return p2;
    end if;
  else
   return greatest(p1,greatest(p2,p3));
  end if;
end;");
      
      oracle_failsafe_ddl("create or replace function rand
return number 
as
begin
  return dbms_random.random;
end;");
      
      oracle_failsafe_ddl("create or replace function substring(p_s varchar2, p_start number, p_length number default null)
return varchar2
as
begin

   if p_length is null then
     return substr(p_s,p_start);
   else
     return substr(p_s,p_start,p_length);
   end if;
   
end;");

  return true;
}

function oracle_failsafe_ddl($ddl) {
     oracle_query("begin execute immediate '$ddl'; exception when others then null; end;");
}
